import{d as _,_ as m,W as y,b as k,e as s,c as e,f as o,g as d,E as n,o as t,k as c,aR as v,a as r,aN as g,q as C,aS as x,x as h}from"./index-BrFqolpA.js";import{u as B}from"./usePageTitle-CsnULHHx.js";const S=_({__name:"ArtifactKey",setup(K){const u=m(),l=y("artifactKey"),i=k(u.artifacts.getArtifactCollection,[l]),a=s(()=>i.response),f=s(()=>a.value?`${n.info.artifact}: ${a.value.key}`:n.info.artifact);return B(f),(b,A)=>{const p=d("p-layout-default");return t(),e(p,{class:"artifact"},{header:o(()=>[a.value?(t(),e(c(v),{key:0,artifact:a.value},null,8,["artifact"])):r("",!0)]),default:o(()=>[a.value?(t(),e(c(g),{key:0,artifact:a.value},null,8,["artifact"])):r("",!0),a.value?(t(),C(h,{key:1},[a.value.key?(t(),e(c(x),{key:0,"artifact-key":a.value.key},null,8,["artifact-key"])):r("",!0)],64)):r("",!0)]),_:1})}}});export{S as default};
//# sourceMappingURL=ArtifactKey-CQe3gUwn.js.map
