import{d as k,_,W as l,J as c,c as w,f as r,g as d,o as i,j as n,k as o,d9 as f,da as P}from"./index-BrFqolpA.js";import{u as Q}from"./usePageTitle-CsnULHHx.js";const x=k({__name:"WorkPoolQueueEdit",async setup(N){let e,u;const s=_(),a=l("workPoolName"),t=l("workPoolQueueName"),p=([e,u]=c(()=>s.workPoolQueues.getWorkPoolQueueByName(a.value,t.value)),e=await e,u(),e);return Q("Edit Work Pool Queue"),(y,W)=>{const m=d("p-layout-default");return i(),w(m,null,{header:r(()=>[n(o(f),{"work-pool-name":o(a),"work-pool-queue-name":o(t)},null,8,["work-pool-name","work-pool-queue-name"])]),default:r(()=>[n(o(P),{"work-pool-name":o(a),"work-pool-queue":o(p)},null,8,["work-pool-name","work-pool-queue"])]),_:1})}}});export{x as default};
//# sourceMappingURL=WorkPoolQueueEdit-faNYfDM5.js.map
