.automation-wizard-action{display:grid;gap:.25rem}.automation-wizard-action__header{display:flex;align-items:center;justify-content:space-between}.automation-wizard-action__heading{font-weight:700}.automation-wizard-step-actions{display:grid;gap:2rem}.automations-wizard-step-actions__add{justify-self:start}.automation-trigger-json-input__json-input{min-height:400px}.automation-wizard-step-trigger__mode-switcher{margin-left:auto;margin-right:auto}
