import{d as k,_ as y,W as i,e as o,bK as b,c as _,f as l,a as m,g as f,o as T,j as t,k as c,cq as d,cr as g}from"./index-BrFqolpA.js";import{u as v}from"./usePageTitle-CsnULHHx.js";const x=k({__name:"BlocksCatalogView",setup(B){const s=y(),a=i("blockTypeSlug"),n=o(()=>a.value?[a.value]:null),u=b(s.blockTypes.getBlockTypeBySlug,n),e=o(()=>u.response),p=o(()=>e.value?`Block Type: ${e.value.name}`:null);return v(p),(C,S)=>{const r=f("p-layout-default");return e.value?(T(),_(r,{key:0,class:"blocks-catalog-view"},{header:l(()=>[t(c(d),{"block-type":e.value},null,8,["block-type"])]),default:l(()=>[t(c(g),{"block-type":e.value},null,8,["block-type"])]),_:1})):m("",!0)}}});export{x as default};
//# sourceMappingURL=BlocksCatalogView-AkxsA4ky.js.map
