import{d as e,c as o,k as a,o as t,de as r}from"./index-BrFqolpA.js";import{u as s}from"./usePageTitle-CsnULHHx.js";const m=e({__name:"404",setup(n){return s("Page Not Found"),(c,p)=>(t(),o(a(r)))}});export{m as default};
//# sourceMappingURL=404-CqvlC-Oy.js.map
