import{d as h,_ as w,W as g,e as u,ao as D,bx as I,b as R,c,f as n,g as l,u as S,aO as T,o as a,k as t,cJ as B,a as r,cK as m,j as N,cL as O,a6 as V,bo as W}from"./index-BrFqolpA.js";import{u as j}from"./usePageTitle-CsnULHHx.js";const K=h({__name:"ConcurrencyLimit",setup(A){const p=w(),y=g("concurrencyLimitId"),v=S(),d=u(()=>[{label:"Details",hidden:T.xl},{label:"Active Task Runs"}]),s=D("tab","Details"),{tabs:_}=I(d,s),b={interval:3e5},L=R(p.concurrencyLimits.getConcurrencyLimit,[y.value],b),e=u(()=>L.response);function f(){v.push(W.concurrencyLimits())}const k=u(()=>e.value?`Concurrency Limit: ${e.value.tag}`:"Concurrency Limit");return j(k),(G,i)=>{const C=l("p-tabs"),x=l("p-layout-well");return a(),c(x,{class:"concurrencyLimit"},{header:n(()=>[e.value?(a(),c(t(B),{key:0,"concurrency-limit":e.value,onDelete:f},null,8,["concurrency-limit"])):r("",!0)]),well:n(()=>[e.value?(a(),c(t(m),{key:0,alternate:"","concurrency-limit":e.value},null,8,["concurrency-limit"])):r("",!0)]),default:n(()=>[N(C,{selected:t(s),"onUpdate:selected":i[0]||(i[0]=o=>V(s)?s.value=o:null),tabs:t(_)},{details:n(()=>[e.value?(a(),c(t(m),{key:0,"concurrency-limit":e.value},null,8,["concurrency-limit"])):r("",!0)]),"active-task-runs":n(()=>{var o;return[(o=e.value)!=null&&o.activeSlots?(a(),c(t(O),{key:0,"active-slots":e.value.activeSlots},null,8,["active-slots"])):r("",!0)]}),_:1},8,["selected","tabs"])]),_:1})}}});export{K as default};
//# sourceMappingURL=ConcurrencyLimit-Cn4a82w2.js.map
