function s(n,e){if(e.length<n)throw new TypeError(n+" argument required, but only "+e.length+" present")}var r=1e3,o=3600,t=o*24,a=t*7;function c(n){return s(1,arguments),n*r}export{t as a,a as b,c as s};
//# sourceMappingURL=index-0yGGjo7r.js.map
