import{d,_,b as i,e as o,c as a,f as n,g as y,o as t,j as f,k as s,b_ as k,q as b,b$ as g,bZ as v,x,a as D}from"./index-BrFqolpA.js";import{u as h}from"./usePageTitle-CsnULHHx.js";const j=d({__name:"Deployments",setup(C){const l=_(),c={interval:3e4},e=i(l.deployments.getDeployments,[{}],c),r=o(()=>e.response??[]),p=o(()=>e.executed&&r.value.length===0),m=o(()=>e.executed);return h("Deployments"),(B,V)=>{const u=y("p-layout-default");return t(),a(u,{class:"deployments"},{header:n(()=>[f(s(k))]),default:n(()=>[m.value?(t(),b(x,{key:0},[p.value?(t(),a(s(g),{key:0})):(t(),a(s(v),{key:1,onDelete:s(e).refresh},null,8,["onDelete"]))],64)):D("",!0)]),_:1})}}});export{j as default};
//# sourceMappingURL=Deployments-DNfOidkm.js.map
