import{d as u,s,w as c,c as l,r as m,a as p,O as i,h as f,i as h,o as v,L as d}from"./index-BrFqolpA.js";const r=a=>{const{value:e,isCustom:o}=i();if(f(a.query)&&o.value)try{const t=h.map("SavedSearchFilter",e.value,"LocationQuery");return{...a,query:t}}catch(t){console.error(t)}return!0},R=u({beforeRouteEnter:r,beforeRouteUpdate:r,__name:"RunsPageWithDefaultFilter",props:{component:{type:Function}},setup(a){const e=a,o=s(null);function t(n){return d(n)}return c(e.component,()=>{t(e.component)?e.component().then(n=>{o.value=n.default}):o.value=e.component},{immediate:!0}),(n,y)=>o.value!==null?(v(),l(m(o.value),{key:0})):p("",!0)}});export{R as default};
//# sourceMappingURL=RunsPageWithDefaultFilter-DjwU_mrc-B6Q6AfnG.js.map
