import{d as y,bU as S,e as t,aV as W,aW as B,i as b,o as a,q as C,l as D,j as n,k as e,bV as N,a$ as T,_ as X,W as $,ao as j,b as U,by as Y,bW as Z,c as i,f as o,g as F,u as q,bX as A,a as p,bY as E,bZ as H,bH as M,a6 as O,bo as P}from"./index-BrFqolpA.js";import{s as z,b as G}from"./index-0yGGjo7r.js";import{u as J}from"./usePageTitle-CsnULHHx.js";const K={class:"flow-stats"},L={class:"flow-stats__cards"},Q=y({__name:"FlowStats",props:{flowId:{}},setup(w){const f=w,r={range:{type:"span",seconds:-G},tags:[]},{flowId:u}=S(f),c=t(()=>({flowId:u.value,range:r.range}));W(B,{interval:z(30)});const l=t(()=>b.map("FlowStatsFilter",c.value,"FlowRunsFilter")),d=t(()=>b.map("FlowStatsFilter",c.value,"TaskRunsFilter"));return(_,m)=>(a(),C("div",K,[D("div",L,[n(e(N),{filter:l.value},null,8,["filter"]),n(e(T),{filter:d.value},null,8,["filter"])])]))}}),oe=y({__name:"Flow",setup(w){const f=X(),r=$("flowId"),u=t(()=>[r.value]),c=q(),l=j("tab","Runs"),d=["Runs","Deployments","Details"],_={interval:3e5},m=U(f.flows.getFlow,[r.value],_),s=t(()=>m.response),{filter:k}=Y({flows:{id:u}}),{filter:R}=Z({flows:{id:u}});function g(){c.push(P.flows())}const I=t(()=>s.value?`Flow: ${s.value.name}`:"Flow");return J(I),(ee,v)=>{const x=F("p-tabs"),h=F("p-layout-default");return a(),i(h,{class:"flow"},{header:o(()=>[s.value?(a(),i(e(A),{key:0,flow:s.value,onDelete:g},null,8,["flow"])):p("",!0)]),default:o(()=>[s.value?(a(),i(Q,{key:0,"flow-id":s.value.id},null,8,["flow-id"])):p("",!0),n(x,{selected:e(l),"onUpdate:selected":v[0]||(v[0]=V=>O(l)?l.value=V:null),tabs:d},{details:o(()=>[s.value?(a(),i(e(E),{key:0,flow:s.value},null,8,["flow"])):p("",!0)]),deployments:o(()=>[n(e(H),{filter:e(R),prefix:"deployments"},null,8,["filter"])]),runs:o(()=>[n(e(M),{filter:e(k),selectable:"",prefix:"runs"},null,8,["filter"])]),_:1},8,["selected"])]),_:1})}}});export{oe as default};
//# sourceMappingURL=Flow--NMXoysK.js.map
