import{d as m,_,b as i,e as o,c as s,f as n,g as w,o as t,j as C,k as a,bR as b,q as k,bS as v,bT as x,x as y,a as h}from"./index-BrFqolpA.js";import{u as g}from"./usePageTitle-CsnULHHx.js";const S=m({__name:"Flows",setup(F){const c=_(),l={interval:3e4},e=i(c.flows.getFlowsCount,[{}],l),r=o(()=>e.response??0),u=o(()=>e.executed&&r.value===0),p=o(()=>e.executed),d=()=>{e.refresh()};return g("Flows"),(B,V)=>{const f=w("p-layout-default");return t(),s(f,{class:"flows"},{header:n(()=>[C(a(b))]),default:n(()=>[p.value?(t(),k(y,{key:0},[u.value?(t(),s(a(v),{key:0})):(t(),s(a(x),{key:1,selectable:"",onDelete:d}))],64)):h("",!0)]),_:1})}}});export{S as default};
//# sourceMappingURL=Flows-Cg1K0pyB.js.map
