.automation-card__header{display:flex;align-items:center;justify-content:space-between;gap:.5rem}.automation-card__header-actions{display:flex;align-items:center;gap:.5rem}.automation-card__name{font-size:1.125rem;line-height:1.75rem}.automation-card__description{font-size:.875rem;line-height:1.25rem}.automation-card__label{margin-right:.5rem;font-weight:500}.automations{--virtual-scroller-item-gap: .5rem }
