import{d as w,cM as p,ao as C,c as M,f as a,g as u,o as h,j as l,k as o,K as r,cN as y,a6 as i,cO as R,cP as x,cQ as g}from"./index-BrFqolpA.js";const L=w({__name:"ConcurrencyLimits",setup(v){const{showModal:t,open:b}=p(),{showModal:e,open:m}=p(),_=[{label:"Global"},{label:"Task Run"}],c=C("tab","Global");return(G,s)=>{const d=u("p-button"),k=u("p-tabs"),f=u("p-layout-default");return h(),M(f,{class:"concurrency-limits"},{header:a(()=>[l(o(r),{crumbs:[{text:"Concurrency"}]})]),default:a(()=>[l(k,{selected:o(c),"onUpdate:selected":s[2]||(s[2]=n=>i(c)?c.value=n:null),tabs:_},{global:a(()=>[l(o(r),{size:"lg",crumbs:[{text:"Global Concurrency Limits"}]},{"after-crumbs":a(()=>[l(d,{small:"",icon:"PlusIcon",onClick:o(b)},null,8,["onClick"])]),_:1}),l(o(y),{showModal:o(t),"onUpdate:showModal":s[0]||(s[0]=n=>i(t)?t.value=n:null)},null,8,["showModal"]),l(o(R),{class:"concurrency-limits__global-table"})]),"task-run":a(()=>[l(o(r),{size:"lg",crumbs:[{text:"Task Run Concurrency Limits"}]},{"after-crumbs":a(()=>[l(d,{small:"",icon:"PlusIcon",onClick:o(m)},null,8,["onClick"])]),_:1}),l(o(x),{showModal:o(e),"onUpdate:showModal":s[1]||(s[1]=n=>i(e)?e.value=n:null)},null,8,["showModal"]),l(o(g),{class:"concurrency-limits__task-limits-table"})]),_:1},8,["selected"])]),_:1})}}});export{L as default};
//# sourceMappingURL=ConcurrencyLimits-lijxxlg3.js.map
