import{d as g,_ as k,b as v,e as s,c as n,f as i,g as r,o as a,j as u,k as o,cB as x,z as N,q as y,cC as C,cD as B,x as D,a as b}from"./index-BrFqolpA.js";import{u as V}from"./usePageTitle-CsnULHHx.js";const $=g({__name:"Notifications",setup(w){const l=k(),t=v(l.notifications.getNotifications),c=s(()=>t.response??[]),f=s(()=>t.executed&&c.value.length===0),p=s(()=>t.executed);return V("Notifications"),(P,e)=>{const d=r("p-message"),m=r("p-layout-default");return a(),n(m,{class:"notifications"},{header:i(()=>[u(o(x))]),default:i(()=>[u(d,{info:""},{default:i(()=>e[2]||(e[2]=[N(" Notifications are deprecated and will be migrated in the future. Please use Automations. ")])),_:1}),p.value?(a(),y(D,{key:0},[f.value?(a(),n(o(C),{key:0})):(a(),n(o(B),{key:1,notifications:c.value,onDelete:e[0]||(e[0]=_=>o(t).refresh()),onUpdate:e[1]||(e[1]=_=>o(t).refresh())},null,8,["notifications"]))],64)):b("",!0)]),_:1})}}});export{$ as default};
//# sourceMappingURL=Notifications-oRD5T00p.js.map
