import{d as _,_ as m,b as k,e as t,c as a,f as n,g as i,o,j as f,k as s,cZ as P,q as x,c_ as y,c$ as g,x as h,a as v}from"./index-BrFqolpA.js";import{u as w}from"./usePageTitle-CsnULHHx.js";const N=_({__name:"WorkPools",setup(b){const r=m(),c={interval:3e4},e=k(r.workPools.getWorkPools,[{}],c),l=t(()=>e.response??[]),p=t(()=>e.executed&&l.value.length==0),u=t(()=>e.executed);return w("Work Pools"),(C,B)=>{const d=i("p-layout-default");return o(),a(d,{class:"work-pools"},{header:n(()=>[f(s(P))]),default:n(()=>[u.value?(o(),x(h,{key:0},[p.value?(o(),a(s(y),{key:0})):(o(),a(s(g),{key:1,onUpdate:s(e).refresh},null,8,["onUpdate"]))],64)):v("",!0)]),_:1})}}});export{N as default};
//# sourceMappingURL=WorkPools-DOYYTHY0.js.map
