import{d as v,e as y,o as i,q as u,j as e,f as t,x,v as C,k as n,db as B,t as f,J as w,c as P,g as l,K as V,dc as j,dd as d,a6 as A}from"./index-BrFqolpA.js";import{u as K}from"./usePageTitle-CsnULHHx.js";import{u as M}from"./usePrefectApi-Q0V-aAwL.js";const T={class:"settings-block"},q=v({__name:"SettingsCodeBlock",props:{engineSettings:{}},setup(g){const s=g,o=y(()=>Object.entries(s.engineSettings));return(p,c)=>(i(),u("div",T,[e(n(B),{multiline:""},{default:t(()=>[(i(!0),u(x,null,C(o.value,(a,r)=>(i(),u("div",{key:r,class:"settings-block--code-line"},f(a[0])+": "+f(a[1]),1))),128))]),_:1})]))}}),J=v({__name:"Settings",async setup(g){let s,o;const p=[{text:"Settings"}],c=M(),[a,r]=([s,o]=w(()=>Promise.all([c.admin.getSettings(),c.admin.getVersion()])),s=await s,o(),s);return K("Settings"),(D,m)=>{const b=l("p-key-value"),S=l("p-theme-toggle"),_=l("p-label"),k=l("p-layout-default");return i(),P(k,{class:"settings"},{header:t(()=>[e(n(V),{crumbs:p},{actions:t(()=>[e(b,{class:"settings__version",label:"Version",value:n(r),alternate:""},null,8,["value"])]),_:1})]),default:t(()=>[e(_,{label:"Theme"},{default:t(()=>[e(S)]),_:1}),e(_,{label:"Color Mode",class:"settings__color-mode"},{default:t(()=>[e(n(j),{selected:n(d),"onUpdate:selected":m[0]||(m[0]=h=>A(d)?d.value=h:null)},null,8,["selected"])]),_:1}),e(_,{label:"Server Settings"},{default:t(()=>[e(q,{class:"settings__code-block","engine-settings":n(a)},null,8,["engine-settings"])]),_:1})]),_:1})}}});export{J as default};
//# sourceMappingURL=Settings-BxsN5J1e.js.map
