import{d as p,_ as c,W as u,J as k,c as m,f as l,g as d,o as w,j as r,k as e,d2 as f,d3 as i}from"./index-BrFqolpA.js";import{u as P}from"./usePageTitle-CsnULHHx.js";const N=p({__name:"WorkPoolEdit",async setup(y){let o,a;const s=c(),n=u("workPoolName"),t=([o,a]=k(()=>s.workPools.getWorkPoolByName(n.value)),o=await o,a(),o);return P("Create Work Pool"),(C,W)=>{const _=d("p-layout-default");return w(),m(_,null,{header:l(()=>[r(e(f),{"work-pool":e(t)},null,8,["work-pool"])]),default:l(()=>[r(e(i),{"work-pool":e(t)},null,8,["work-pool"])]),_:1})}}});export{N as default};
//# sourceMappingURL=WorkPoolEdit-Dl468SJu.js.map
