import{d as k,_ as f,af as _,e as o,cn as y,b as m,c as d,f as s,g,o as v,j as t,k as l,co as B,cp as C}from"./index-BrFqolpA.js";import{u as T}from"./usePageTitle-CsnULHHx.js";const q=k({__name:"BlocksCatalog",setup(x){const c=f(),a=_(null),p=o(()=>a.value?[a.value]:[]),{filter:n}=y({blockSchemas:{blockCapabilities:p}}),i=m(c.blockTypes.getBlockTypes,[n]),u=o(()=>i.response??[]);return T("Blocks Catalog"),(h,e)=>{const r=g("p-layout-default");return v(),d(r,{class:"blocks-catalog"},{header:s(()=>[t(l(B))]),default:s(()=>[t(l(C),{capability:a.value,"onUpdate:capability":e[0]||(e[0]=b=>a.value=b),"block-types":u.value},null,8,["capability","block-types"])]),_:1})}}});export{q as default};
//# sourceMappingURL=BlocksCatalog-qnd9HlwA.js.map
