import{d as B,_ as I,W as R,b as S,e as p,af as T,ao as V,c,f as e,g as n,E as _,aK as $,o as l,k as t,aL as h,a as i,aM as d,q,j as s,aN as E,aO as P,aP as v,z as W,t as j,aQ as m,a6 as z}from"./index-BrFqolpA.js";import{u as F}from"./usePageTitle-CsnULHHx.js";const H={key:0},O=B({__name:"Artifact",setup(J){const b=I(),k=R("artifactId"),w=S(b.artifacts.getArtifact,[k]),a=p(()=>w.response),r=T(!1),y=[{label:"Artifact"},{label:"Details"},{label:"Raw"}],u=V("tab","Artifact"),g=p(()=>a.value?`${_.info.artifact}: ${a.value.key??$(a.value.type)}`:_.info.artifact);return F(g),(K,o)=>{const x=n("p-divider"),C=n("p-button"),A=n("p-content"),D=n("p-tabs"),N=n("p-layout-well");return l(),c(N,{class:"artifact"},{header:e(()=>[a.value?(l(),c(t(h),{key:0,artifact:a.value},null,8,["artifact"])):i("",!0)]),well:e(()=>[a.value?(l(),c(t(d),{key:0,artifact:a.value,alternate:""},null,8,["artifact"])):i("",!0)]),default:e(()=>[a.value?(l(),q("section",H,[s(t(E),{artifact:a.value},null,8,["artifact"]),s(x),t(P).xl?(l(),c(A,{key:0},{default:e(()=>[s(t(v),{artifact:a.value},null,8,["artifact"]),s(C,{class:"artifact__raw-data-button",small:"",onClick:o[0]||(o[0]=f=>r.value=!r.value)},{default:e(()=>[W(j(r.value?"Hide":"Show")+" raw data ",1)]),_:1}),r.value?(l(),c(t(m),{key:0,artifact:a.value},null,8,["artifact"])):i("",!0)]),_:1})):(l(),c(D,{key:1,selected:t(u),"onUpdate:selected":o[1]||(o[1]=f=>z(u)?u.value=f:null),tabs:y},{artifact:e(()=>[s(t(v),{artifact:a.value},null,8,["artifact"])]),details:e(()=>[s(t(d),{artifact:a.value},null,8,["artifact"])]),raw:e(()=>[s(t(m),{artifact:a.value},null,8,["artifact"])]),_:1},8,["selected"]))])):i("",!0)]),_:1})}}});export{O as default};
//# sourceMappingURL=Artifact-OzLudexG.js.map
