import{d as f,_ as x,B as b,W as c,e as r,aG as y,J as B,k as e,c as C,f as n,a as h,g as k,aH as E,o as I,j as s,aI as g,K as w,aJ as D}from"./index-BrFqolpA.js";import{u as W}from"./usePageTitle-CsnULHHx.js";const V=f({__name:"Event",async setup(F){let a,o;const u=x(),v=b(),l=c("eventDate"),_=c("eventId"),d=r(()=>E(l.value)),m=y({startDate:d,eventId:[_.value]}),t=([a,o]=B(()=>u.events.getFirstEvent(m.value)),a=await a,o(),a),p=r(()=>[{text:"Event Feed",to:v.events()},{text:t.eventLabel}]);return W(`Event: ${t.eventLabel}`),(J,K)=>{const i=k("p-layout-default");return e(t)?(I(),C(i,{key:0,class:"event"},{header:n(()=>[s(e(w),{crumbs:p.value},{actions:n(()=>[s(e(g),{event:e(t)},null,8,["event"])]),_:1},8,["crumbs"])]),default:n(()=>[s(e(D),{event:e(t)},null,8,["event"])]),_:1})):h("",!0)}}});export{V as default};
//# sourceMappingURL=Event-CLQVKPLW.js.map
