import{d as o,c as r,f as a,g as c,o as n,j as e,k as t,aT as f,aU as _}from"./index-BrFqolpA.js";import{u as p}from"./usePageTitle-CsnULHHx.js";const h=o({__name:"Artifacts",setup(u){return p("Artifacts"),(l,d)=>{const s=c("p-layout-default");return n(),r(s,{class:"artifacts"},{header:a(()=>[e(t(f))]),default:a(()=>[e(t(_))]),_:1})}}});export{h as default};
//# sourceMappingURL=Artifacts--KK5OVeD.js.map
