import{d as _,af as d,_ as m,e as o,E as v,c as r,f as c,g as b,b as h,o as t,j as k,k as n,cR as y,q as x,cS as C,cT as g,x as V,a as B}from"./index-BrFqolpA.js";import{u as I}from"./usePageTitle-CsnULHHx.js";const j=_({__name:"Variables",setup(S){const l=d(),a=()=>{var e;s.value.refresh(),(e=l.value)==null||e.refreshSubscriptions()},u=m(),s=o(()=>h(u.variables.getVariables)),p=o(()=>{var e;return s.value.executed&&((e=s.value.response)==null?void 0:e.length)===0}),f=o(()=>s.value.executed);return I(v.info.variables),(e,E)=>{const i=b("p-layout-default");return t(),r(i,{class:"variables"},{header:c(()=>[k(n(y),{onCreate:a})]),default:c(()=>[f.value?(t(),x(V,{key:0},[p.value?(t(),r(n(C),{key:0,onCreate:a})):(t(),r(n(g),{key:1,ref_key:"table",ref:l,onDelete:a,onUpdate:a},null,512))],64)):B("",!0)]),_:1})}}});export{j as default};
//# sourceMappingURL=Variables-Npu1zPiJ.js.map
