import{d as p,_,c as l,f as o,g as m,o as d,j as e,k as n,cE as C,cF as h,cc as c,bo as i,P as y}from"./index-BrFqolpA.js";import{u as N}from"./usePageTitle-CsnULHHx.js";const w=p({__name:"NotificationCreate",setup(b){const r=_();async function s(t){try{await r.notifications.createNotification(t),c.push(i.notifications())}catch(a){y("Error creating notification","error"),console.warn(a)}}function u(){c.push(i.notifications())}return N("Create Notification"),(t,a)=>{const f=m("p-layout-default");return d(),l(f,{class:"notification-create"},{header:o(()=>[e(n(C))]),default:o(()=>[e(n(h),{action:"Create",onSubmit:s,onCancel:u})]),_:1})}}});export{w as default};
//# sourceMappingURL=NotificationCreate-z_OQi0fN.js.map
