import{d as r,c as s,f as e,g as n,o as l,j as o,k as a,d0 as p,d1 as u}from"./index-BrFqolpA.js";import{u as _}from"./usePageTitle-CsnULHHx.js";const k=r({__name:"WorkPoolCreate",setup(c){return _("Create Work Pool"),(d,f)=>{const t=n("p-layout-default");return l(),s(t,null,{header:e(()=>[o(a(p))]),default:e(()=>[o(a(u))]),_:1})}}});export{k as default};
//# sourceMappingURL=WorkPoolCreate-DCv7juoP.js.map
