import{d as p,_ as m,b as _,e as a,c as s,f as c,g as d,o as t,j as k,k as o,ck as f,q as i,cl as x,cm as y,x as B,a as C}from"./index-BrFqolpA.js";import{u as D}from"./usePageTitle-CsnULHHx.js";const V=p({__name:"Blocks",setup(b){const n=m(),e=_(n.blockDocuments.getBlockDocumentsCount),l=a(()=>e.executed&&e.response==0),r=a(()=>e.executed);return D("Blocks"),(g,h)=>{const u=d("p-layout-default");return t(),s(u,{class:"blocks"},{header:c(()=>[k(o(f))]),default:c(()=>[r.value?(t(),i(B,{key:0},[l.value?(t(),s(o(x),{key:0})):(t(),s(o(y),{key:1,onDelete:o(e).refresh},null,8,["onDelete"]))],64)):C("",!0)]),_:1})}}});export{V as default};
//# sourceMappingURL=Blocks-DdfNqZog.js.map
