import{d as b,_ as d,W as i,e as o,bK as _,c as f,f as c,a as v,g as D,u as B,o as g,j as u,k as n,cx as h,cy as x,bo as y}from"./index-BrFqolpA.js";import{u as C}from"./usePageTitle-CsnULHHx.js";const T=b({__name:"BlockView",setup(w){const s=d(),a=B(),t=i("blockDocumentId"),l=o(()=>t.value?[t.value]:null),r=_(s.blockDocuments.getBlockDocument,l),e=o(()=>r.response),m=()=>{a.push(y.blocks())},k=o(()=>e.value?`Block: ${e.value.name}`:"Block");return C(k),(N,V)=>{const p=D("p-layout-default");return e.value?(g(),f(p,{key:0,class:"block-view"},{header:c(()=>[u(n(h),{"block-document":e.value,onDelete:m},null,8,["block-document"])]),default:c(()=>[u(n(x),{"block-document":e.value},null,8,["block-document"])]),_:1})):v("",!0)}}});export{T as default};
//# sourceMappingURL=BlockView-DoNcUSpD.js.map
